package main;

import io.DienstenParser;
import io.RoosterIndelingWriter;
import io.RoosterParser;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import types.Dienst;
import types.DienstRooster;
import types.Diensten;
import types.Rooster;
import types.RoosterIndeling;
import types.Roosters;
import datastructures.BeperkteQueue;
import datastructures.TaboeElement;
import datastructures.ZoekBoom;

public class Planner
{
	private Roosters roosters;
	private Diensten diensten;
	private final int grootteTaboeLijst = 80;
	private final int afstandTotLaatsteVerbetering = 3000;
	private BeperkteQueue <TaboeElement> taboeLijst;

	
	public Planner(Roosters roosters, Diensten diensten)
	{
		this.roosters = roosters;
		this.diensten = diensten;
	}

	/**
	 * 
	 * @param dienstA geeft aan eerste component van omwisseling (nu op [regelnr, dagnr])
	 * @param dienstB geeft aan tweede component van omwisseling (nu op [regelnr, dagnr])
	 * @param toewijzing
	 * @return
	 */
	private double omwisselingsWaarde(int[] dienstA, int[] dienstB, DienstRooster toewijzing)
	{
		return toewijzing.berekenKostenOPTintern(dienstA[1], dienstA[0], dienstB[1], dienstB[0]);
	}
		
	private double omwisselingsWaarde(int[] is, DienstRooster toewijzing1, int[] is2, DienstRooster toewijzing2) 
	{
		return toewijzing1.berekenKostenOPTextern(is[1], is[0], toewijzing2.getDienst(is2[0], is2[1]));
	}
	
	/**
	 * Wisselt de diensten A en B om in de gegeven dienstentoewijzing.
	 * @param dienstA geeft aan eerste component van omwisseling (nu op [regelnr, dagnr])
	 * @param dienstB geeft aan tweede component van omwisseling (nu op [regelnr, dagnr])
	 * @param toewijzing
	 */
	private void wisselOm(int[] dienstA, int[] dienstB, DienstRooster toewijzing)
	{
		// maak een backup van dienst a
		Dienst backupA = toewijzing.getDienst(dienstA[0], dienstA[1]);
		// a gaat naar b wijzen
		toewijzing.addDienst(dienstA[0], dienstA[1], toewijzing.getDienst(dienstB[0], dienstB[1]));
		// en b gaat naar a wijzen
		toewijzing.addDienst(dienstB[0], dienstB[1], backupA);
	}
	
	/**
	 * Wisselt de diensten A en B om in de gegeven dienstentoewijzing.
	 * @param dienstA geeft aan eerste component van omwisseling (nu op [regelnr, dagnr])
	 * @param toewijzing1 geeft het dienstrooster horende bij dienstA
	 * @param dienstB geeft aan tweede component van omwisseling (nu op [regelnr, dagnr])
	 * @param toewijzing2 geeft het dienstrooster horende bij dienstB
	 */
	private void wisselOm(int[] dienstA, DienstRooster toewijzing1, int[] dienstB, DienstRooster toewijzing2) 
	{
		// maak een backup van dienst a
		Dienst backupA = toewijzing1.getDienst(dienstA[0], dienstA[1]);
		// a gaat naar b wijzen
		toewijzing1.addDienst(dienstA[0], dienstA[1], toewijzing2.getDienst(dienstB[0], dienstB[1]));
		// en b gaat naar a wijzen
		toewijzing2.addDienst(dienstB[0], dienstB[1], backupA);		
	}
	
	private RoosterIndeling search(RoosterIndeling roosterIndeling)
	{	
	    double waardeBesteOplossing = Double.MAX_VALUE;
		RoosterIndeling besteIndeling = null;
		waardeBesteOplossing = roosterIndeling.getKosten();
		besteIndeling = roosterIndeling;
		
		taboeLijst = new BeperkteQueue<TaboeElement>(grootteTaboeLijst);
		int[] roosternrs = roosterIndeling.getRoosternrs();
		// doe een omwisseling in elk rooster
		int besteOplossingAangepast = 0;
		int j=0;

		while(besteOplossingAangepast<afstandTotLaatsteVerbetering)
		{	
			for(int i=0; i<roosternrs.length;i++)
			{
				DienstRooster toewijzing = roosterIndeling.getRoosterIndeling(roosternrs[i]);
				double oudeWaarde = toewijzing.getKosten();
				double waarde = localSearchStap(toewijzing);
				if(waarde<oudeWaarde)
				{
					waardeBesteOplossing = roosterIndeling.getKosten();
					besteIndeling = roosterIndeling;
					besteOplossingAangepast = 0;
					
				}
				else
					besteOplossingAangepast++;
				int randomRooster = (int)(Math.random()*(float)roosternrs.length);
				DienstRooster toewijzing2 = roosterIndeling.getRoosterIndeling(roosternrs[randomRooster]);
				double oudeWaarde2 = toewijzing2.getKosten();
				double waarde2 = localSearchStap(toewijzing, toewijzing2);
				if(waarde2<oudeWaarde2)
				{
					waardeBesteOplossing = roosterIndeling.getKosten();
					besteIndeling = roosterIndeling;
					besteOplossingAangepast = 0;
				}
				else
					besteOplossingAangepast++;
			}
			if(j%10 == 0)
				System.out.println("j " + j + " " + waardeBesteOplossing);
			j++;
		}
		return besteIndeling;
	}

	private double localSearchStap(DienstRooster toewijzing)
	{
		//bereken buren bij 2-OPT (of 1-OPT)
		int cycleLengte = toewijzing.getAantalRegels();
		double besteTotNuToe = Double.MAX_VALUE;
		int [][]beste = new int[2][2];
		boolean betereGevonden = false;
		int d = (int)(Math.random()*7.0);
		int t = (int)(Math.random()*cycleLengte);
		for(int dag=d; dag<7 && !betereGevonden; dag++)
		{
			for(int t1=t; t1<cycleLengte && !betereGevonden; t1++)
			{
				for(int t2= t1+1; t2<cycleLengte && !betereGevonden; t2++)
				{
						double waarde = 
							omwisselingsWaarde(new int[] {t1, dag}, new int[] {t2, dag}, toewijzing);
						// verandering is verbering
						if(waarde < toewijzing.getKosten())
						{
							wisselOm(new int[]{t1, dag}, new int[]{t2,dag},toewijzing);
							TaboeElement wissel1 = new TaboeElement(toewijzing.getRoosterNr(), t1, dag,toewijzing.getDienst(t1, dag));
							TaboeElement wissel2 = new TaboeElement(toewijzing.getRoosterNr(), t2, dag,toewijzing.getDienst(t2, dag));
							// als taboe draai dan je beslissing terug						
							if(taboe(wissel1, wissel2))
								wisselOm(new int[]{t2, dag}, new int[]{t1,dag},toewijzing);
							// iteratie voltooid
							else
							{
								betereGevonden= true;
								// de elementen zijn net omgedraaid, oude toewijzing in taboelijst
								TaboeElement taboeElement1 = new TaboeElement(toewijzing.getRoosterNr(), t1, dag,toewijzing.getDienst(t2, dag));
								TaboeElement taboeElement2 = new TaboeElement(toewijzing.getRoosterNr(), t2, dag,toewijzing.getDienst(t1, dag));
								adjustTaboeLijst(taboeElement1, taboeElement2);
								return waarde;
							}					
						}
						else if(waarde<besteTotNuToe)
						{
							besteTotNuToe = waarde;
							beste[0][0] = t1; beste[0][1] = dag;
							beste[1][0] = t2; beste[1][1] = dag;
							
						}
				}
			}
		}
		//geen betere gevonden, dan proberen om de beste verslechtering te doen.
		int t1 = beste[0][0]; int t2 = beste[1][0]; int dag = beste[0][1];
			wisselOm(new int[]{t1, dag}, new int[]{t2,dag},toewijzing);
			TaboeElement wissel1 = new TaboeElement(toewijzing.getRoosterNr(), t1, dag,toewijzing.getDienst(t1, dag));
			TaboeElement wissel2 = new TaboeElement(toewijzing.getRoosterNr(), t2, dag,toewijzing.getDienst(t2, dag));
			// als taboe draai dan je beslissing terug						
			if(taboe(wissel1, wissel2))
				wisselOm(new int[]{t2, dag}, new int[]{t1,dag},toewijzing);
			// iteratie voltooid
			return besteTotNuToe;
	}
	
	private double localSearchStap(DienstRooster toewijzing1, DienstRooster toewijzing2) 
	{
		int cycleLengte1 = toewijzing1.getAantalRegels();
		int cycleLengte2 = toewijzing2.getAantalRegels();
		int regel1 = (int)(Math.random()*(double)cycleLengte1);
		int regel2 = (int)(Math.random()*(double)cycleLengte2);
		int d    = (int)(Math.random()*6.0);
		boolean betereGevonden = false;
		double besteTotNuToe = Integer.MAX_VALUE;
		int [][]beste = new int[2][2];
		
		//vanaf een random plek, loop 2 roosters door
		for(int dag=d; dag<7 && !betereGevonden; dag++)
		{
			for(int t1=regel1; t1<cycleLengte1 && !betereGevonden; t1++)
			{
				for(int t2 = regel2; t2<cycleLengte2 && !betereGevonden; t2++)
				{
					double waarde = omwisselingsWaarde(new int[] {t1, dag}, toewijzing1,
							new int[] {t2, dag}, toewijzing2);
					//verandering is verbetering
					if(waarde<toewijzing1.getKosten()+toewijzing2.getKosten())
					{
						wisselOm(new int[]{t1, dag},toewijzing1, new int[]{t2,dag},toewijzing2);
						TaboeElement wissel1 = new TaboeElement(toewijzing1.getRoosterNr(), t1, dag,toewijzing1.getDienst(t1, dag));
						TaboeElement wissel2 = new TaboeElement(toewijzing2.getRoosterNr(), t2, dag,toewijzing2.getDienst(t2, dag));
						// als taboe draai dan je beslissing terug						
						if(taboe(wissel1, wissel2))
							wisselOm(new int[]{t2, dag},toewijzing2, new int[]{t1,dag},toewijzing1);
						// iteratie voltooid
						else
						{
							betereGevonden= true;
							// de elementen zijn net omgedraaid, oude toewijzing in taboelijst
							TaboeElement taboeElement1 = new TaboeElement(toewijzing1.getRoosterNr(),t1, dag,toewijzing2.getDienst(t2, dag));
							TaboeElement taboeElement2 = new TaboeElement(toewijzing2.getRoosterNr(),t2, dag,toewijzing1.getDienst(t1, dag));
							adjustTaboeLijst(taboeElement1, taboeElement2);
							return toewijzing1.getKosten()+toewijzing2.getKosten();
						}	
					}
					else if(waarde<besteTotNuToe)
					{
						besteTotNuToe = waarde;
						beste[0][0] = t1; beste[0][1] = dag;
						beste[1][0] = t2; beste[1][1] = dag;
						
					}
				}
			}
		}
		//geen betere gevonden, dan proberen om de beste verslechtering te doen.
		int t1 = beste[0][0]; int t2 = beste[1][0]; int dag = beste[0][1];
		wisselOm(new int[]{t1, dag},toewijzing1, new int[]{t2,dag},toewijzing2);
		TaboeElement wissel1 = new TaboeElement(toewijzing1.getRoosterNr(), t1, dag,toewijzing1.getDienst(t1, dag));
		TaboeElement wissel2 = new TaboeElement(toewijzing2.getRoosterNr(), t2, dag,toewijzing2.getDienst(t2, dag));
		// als taboe draai dan je beslissing terug						
		if(taboe(wissel1, wissel2))
			wisselOm(new int[]{t2, dag},toewijzing2, new int[]{t1,dag},toewijzing1);
		return toewijzing1.getKosten()+toewijzing2.getKosten();
	}


	/**
	 * 
	 * @param wissel1 geeft het 1e element wat omgewisseld wordt
	 * @param wissel2 geeft het 2e element wat omgewisseld wordt
	 * @return geeft aan of deze wisseling taboe is of niet
	 */
	private boolean taboe(TaboeElement wissel1, TaboeElement wissel2) 
	{
		for(TaboeElement taboeElement: taboeLijst.getQueue())
		{
			if(wissel1.gelijk(taboeElement)||wissel2.gelijk(taboeElement))
				return true;
		}
		return false;
	}

	/**
	 * 
	 * @param taboeElement1 geeft het 1e element wat aan de taboelijst wordt toegevoegd
	 * @param taboeElement2 geeft het 2e element wat aan de taboelijst wordt toegevoegd
	 */
	private void adjustTaboeLijst(TaboeElement taboeElement1,TaboeElement taboeElement2) 
	{
		taboeLijst.add(taboeElement1);
		taboeLijst.add(taboeElement2);
	}


	/**
	 * Maakt een eerste roosterindeling aan, zonder teveel op beperkingen te letten
	 * @return een initieel rooster
	 */
	private RoosterIndeling initieel()
	{
		ZoekBoom<Integer, Dienst> dienstTree = diensten.getDagNummerNaarDienst();
		RoosterIndeling roosterIndeling = new RoosterIndeling();
		int [] roosternrs = new int[roosters.getRoosters().size()];
		int i=0;
		for (Rooster rooster : roosters.getRoosters())
		{
			DienstRooster roosterToewijzing = new DienstRooster(rooster.getRegels(), rooster);
			
			int werkDagen =0;
			for(int regel=0; regel<rooster.getRegels(); regel++)
			{
				for(int dag=0; dag<7; dag++)
				{
					if(werkDagen < 5)
					{
						Dienst dienst = dienstTree.remove(dag);
						if(dienst != null)
						{
							roosterToewijzing.addDienst(regel, dag, dienst);
							werkDagen++;
						}
						else
							roosterToewijzing.addDienst(regel, dag, new Dienst(dag));
					}
					else
					{
						roosterToewijzing.addDienst(regel, dag, new Dienst(dag));
						werkDagen = 0;
					}
				}
			}
			roosterToewijzing.setRoosterNr(rooster.getRoosterNr());
			roosternrs[i]= rooster.getRoosterNr();
			i++;
			
			roosterIndeling.addRoosterIndeling(rooster.getRoosterNr(), roosterToewijzing);
		}
		roosterIndeling.setRoosternrs(roosternrs);
		return roosterIndeling;
	}

	/**
	 * Maakt een planner aan, die gaat proberen voor de gegeven bestanden een
	 * planning te maken.
	 * 
	 * @param args als eerste verplichte argument het pad naar het roosterbestand
	 * en als tweede verplichte argument het pad naar het dienstenbestand.
	 */
	

	public static void main(String[] args)
	{
		// parse de argumenten
		if (args.length >= 2)
		{
			try
			{
				Roosters roosters = new RoosterParser(new File(args[0])).leesRoosters();
				Diensten diensten = new DienstenParser(new File(args[1])).leesDiensten();
				Planner planner = new Planner(roosters, diensten);
				RoosterIndeling roosterIndeling = planner.initieel();
				RoosterIndeling besteIndeling = planner.search(roosterIndeling);
				RoosterIndelingWriter indelingWriter = new RoosterIndelingWriter(new File("output.txt"));
				indelingWriter.schrijf(besteIndeling);
				
			} catch (FileNotFoundException e)
			{
				System.out
						.println("Het opgegeven bestand kon niet worden gevonden.");
			} catch (IOException e)
			{
				System.out
						.println("Het opgegeven bestand kon niet worden gelezen.");
			}
		} else
			System.out.println("U heeft niet genoeg argumenten meegegeven.");
	}
}
