package types;

/**
 * Representeert een roostertoewijzing. Dit is dan een specifiek 
 * cyclisch rooster voor een bepaald diensttype.
 * @author Pieter van Ede
 */
public class DienstRooster 
{
	/**
	 * Eerste dimensie is dag, tweede is regel.
	 */
	private Dienst[][] dienstRooster;
	private RoosterInformatie roosterInformatie;
	private Rooster rooster;	
	private double kosten;
	private int roosterNr;
	
	private final int overUrenKosten = 100;
	private final int extraWerkDagenKosten = 1000;
	private final int kostenLaatVroeg = 6;
	private final int kostenVroegVrijVroeg = 3;
	private final double kostenZaterdagFactor = 0.01;
	private final int kostenZaterdagMacht = 2;
	private final double kostenZondagFactor = 0.01;
	private final int kostenZondagMacht = 2;
	private final double kostenGebrokenDienstenFactor = 0.05;
	private final int kostenGebrokenDienstenMacht = 2;
	private final double kostenLangeDienstenFactor = 0.05;
	private final int kostenLangeDienstenMacht = 2;
	private final double kostenAaneengeslotenRustFactor = -2.5;
	private final int kostenAaneengeslotenRustMacht = 2;
	private final int kostenZaterdagVrijZondagWerken = 50;
	private final int kostenTeWeinigRustTussenDiensten = 10000;
	private final int kostenTeWeinigWekelijkseRust = 70000;
	
	/**
	 * Maakt een dienstRooster aan voor het gegeven aantal regels,
	 * uiteraard wel met zeven dagen.
	 * @param aantalRegels
	 * @param aantalRegels het rooster waarop de diensten hier zijn gepland
	 */
	public DienstRooster(int aantalRegels, Rooster rooster)
	{
		dienstRooster = new Dienst[7][aantalRegels];
		roosterInformatie = new RoosterInformatie(aantalRegels);
		this.rooster = rooster;
		kosten = Integer.MAX_VALUE;
	}
	
	public DienstRooster(Dienst[][] dienstRooster, Rooster rooster) 
	{
		this.dienstRooster = dienstRooster;
		roosterInformatie = new RoosterInformatie(dienstRooster[0].length);
		this.rooster = rooster;
		berekenKosten();
	}
	
	public void berekenKosten()
	{
		kosten = 0;
		
		// bereken enkele hulpconstanten
		int normaleWerkweekLengteInMinuten = (int)(60*rooster.getLengteWerkweek());
		
		// initialiseer alle tellers voor de verschillende kostenposten
		int overMinuten = 0;
		int aantalWerkDagen = 0;
		int dagenTussenZaterdagenInitieel = 0;
		int dagenTussenZaterdagen = 0;
		boolean eersteZaterdag = true;
		int dagenTussenZondagenInitieel = 0;
		int dagenTussenZondagen = 0;
		boolean eersteZondag = true;
		int dagenTussenGebrokenDienstenInitieel = 0;
		int dagenTussenGebrokenDiensten = 0;
		boolean eersteGebrokenDienst = true;
		int dagenTussenLangeDienstenInitieel = 0;
		int dagenTussenLangeDiensten = 0;
		boolean eersteLangeDienst = true;
		
		for(int regelTeller=0; regelTeller<dienstRooster[0].length; regelTeller++)
		{
			int werkweekLengteInMinuten = 0;
			int maximaleAaneengeslotenRustInMinuten = 0;
			
			for(int dagTeller=0; dagTeller<dienstRooster.length; dagTeller++)
			{
				Dienst dienst = dienstRooster[dagTeller][regelTeller];
				
				// update lengte werkweek
				werkweekLengteInMinuten += dienst.getDienstLengte().naarMinuten();
				roosterInformatie.setWerkMinuten(regelTeller, dienst.getDienstLengte().naarMinuten());
				// als geen vrije dag, verhoog het aantal werkdagen
				if(!dienst.isVrij())
					aantalWerkDagen++;
				// bereken of sprake van vroeg-laat of laat-vroeg
				int[] volgendeDag = volgendeDag(new int[] {dagTeller, regelTeller});
				Dienst volgendeDienst = dienstRooster[volgendeDag[0]][volgendeDag[1]];
				if((dienst.getDienstType() == 'V' && volgendeDienst.getDienstType() == 'L')
					|| (dienst.getDienstType() == 'L' && volgendeDienst.getDienstType() == 'V'))
				{
					kosten += kostenLaatVroeg;
				}
				// bereken of sprake van vroeg-vrij-vroeg of laat-vrij-laat
				int[] naVolgendeDag = volgendeDag(volgendeDag);
				Dienst naVolgendeDienst = dienstRooster[naVolgendeDag[0]][naVolgendeDag[1]];
				if((dienst.getDienstType() == 'V' && volgendeDienst.isVrij() && naVolgendeDienst.getDienstType() == 'V')
					|| (dienst.getDienstType() == 'L' && volgendeDienst.isVrij() && naVolgendeDienst.getDienstType() == 'L'))
				{
					kosten += kostenVroegVrijVroeg;
				}
				// bereken de kosten voor het teveel werken op zaterdag of zondag
				if(dagTeller == 6)
				{
					// we zijn op zaterdag
					if(dienst.isVrij())
					{
						if(eersteZaterdag)
						{
							dagenTussenZaterdagenInitieel = dagenTussenZaterdagen;
							eersteZaterdag = false;
						}
						else
						{
							kosten += kostenZaterdagFactor * Math.pow(dagenTussenZaterdagen, kostenZaterdagMacht);
						}
						dagenTussenZaterdagen = 0;
						dagenTussenZondagen++;
					} 
					// op zaterdag gaat de tijd naar volgende zondag altijd door, maar als 
					// je moet werken ook de tijd naar de volgende zaterdag
					else
					{
						dagenTussenZaterdagen++;
						dagenTussenZondagen++;
					}
				}
				else if(dagTeller == 0)
				{
					// we zijn op zondag
					if(dienst.isVrij())
					{
						if(eersteZondag)
						{
							dagenTussenZondagenInitieel = dagenTussenZondagen;
							eersteZondag = false;
						}
						else
						{
							kosten += kostenZondagFactor * Math.pow(dagenTussenZondagen, kostenZondagMacht);
						}
						dagenTussenZondagen = 0;	
						dagenTussenZaterdagen++;
					}
					else
					{
						dagenTussenZaterdagen++;
						dagenTussenZondagen++;
					}
				}
				else
				{
					dagenTussenZondagen++;
					dagenTussenZondagen++;
				}
				// bereken kosten voor teveel gebroken diensten
				if(dienst.getDienstType() == 'G')
				{
					if(eersteGebrokenDienst)
					{
						dagenTussenGebrokenDienstenInitieel = dagenTussenGebrokenDiensten;
						eersteGebrokenDienst = false;
					}
					else
					{
						kosten += kostenGebrokenDienstenFactor * Math.pow(dagenTussenGebrokenDiensten, 
								kostenGebrokenDienstenMacht);
					}
					
					dagenTussenGebrokenDiensten = 0; 
				}
				else
					dagenTussenGebrokenDiensten++;
				// bereken kosten voor teveel lange diensten
				if(dienst.getDienstLengte().naarMinuten() >= 510 /*meer dan 8:30*/)
				{
					if(eersteLangeDienst)
					{
						dagenTussenLangeDienstenInitieel = dagenTussenLangeDiensten;
						eersteLangeDienst = false;
					}
					else
					{
						kosten += kostenLangeDienstenFactor * Math.pow(dagenTussenLangeDiensten, 
								kostenLangeDienstenMacht);
					}
					dagenTussenLangeDiensten = 0; 
				}
				else
					dagenTussenLangeDiensten++;
				// bereken bonus voor lange wekelijkse rust
				if(!dienst.isVrij())
				{
					// tijd tot middernacht
					Tijdstip middernacht = new Tijdstip(24, 00);
					int minutenRust = (middernacht.minus(dienst.getEindTijd())).naarMinuten();
					// loop vooruit om de rust voor de volgende dienst (niet vrije dag) te bepalen
					int[] komendeDag = volgendeDag(new int[] {dagTeller, regelTeller});
					Dienst komendeDienst = dienstRooster[komendeDag[0]][komendeDag[1]];
					while (komendeDienst != null)
					{
						if(komendeDienst.isVrij())
						{
							minutenRust += 1440; // 24*60
							komendeDag = volgendeDag(new int[] {dagTeller, regelTeller});
							komendeDienst = dienstRooster[volgendeDag[0]][volgendeDag[1]];
						}
						else
						{
							minutenRust += (dienst.getBeginTijd().minus(middernacht)).naarMinuten();
							komendeDienst = null;
						}
					}
					maximaleAaneengeslotenRustInMinuten = Math.max(maximaleAaneengeslotenRustInMinuten, minutenRust);
				}
				// bereken boete als je zaterdag vrij bent maar zondag moet werken
				if(dienst.getDagNummer() == 6 && dienst.isVrij())
				{
					if(!volgendeDienst.isVrij())
					{
						kosten += kostenZaterdagVrijZondagWerken;
					}
				}
				// bereken de boete als er te weinig tijd zit tussen twee opvolgende diensten
				if(!dienst.isVrij() && !volgendeDienst.isVrij())
				{
					if((volgendeDienst.getEindTijd().minus(dienst.getEindTijd())).naarMinuten()<660) //11*60
					{
						kosten += kostenTeWeinigRustTussenDiensten;
					}
				}
			}
			// bepaal het aantal over'minuten' per week
			overMinuten += Math.max(0, werkweekLengteInMinuten-normaleWerkweekLengteInMinuten);
			// bereken de kosten voor de langste aaneengesloten rust deze week
			double rustUren = ((float)maximaleAaneengeslotenRustInMinuten)/60.0;
			kosten += kostenAaneengeslotenRustFactor * Math.pow(rustUren, kostenAaneengeslotenRustMacht);
			roosterInformatie.setWekelijkseRust(regelTeller, maximaleAaneengeslotenRustInMinuten);
			
			// bereken boete voor te weinig aaneengesloten rust
			int vorigeRegel;
			if(regelTeller>0)
				vorigeRegel = regelTeller;
			else
				vorigeRegel = 0;
			if(rustUren < 36)
				kosten += kostenTeWeinigWekelijkseRust;
			double vorigeWeekTekort = Math.max(0, 45-roosterInformatie.getWekelijkseRust(vorigeRegel));
			if(vorigeWeekTekort>0 && rustUren-vorigeWeekTekort<45)
				kosten += kostenTeWeinigWekelijkseRust;
		}
		
		// bereken de kosten voor overuren
		float overUren = ((float)overMinuten) / 60;
		kosten += overUrenKosten*overUren;
		// bereken de kosten voor teveel werkdagen
		kosten += Math.max(0, aantalWerkDagen-rooster.getWerkdagen()) * extraWerkDagenKosten;
		roosterInformatie.setWerkdagen(aantalWerkDagen);
		// bereken de kosten voor de tijd tot het eerste gebeurtenis als een lange dienst
		kosten += kostenZaterdagFactor * Math.pow(dagenTussenZaterdagenInitieel + dagenTussenZaterdagen, kostenZaterdagMacht);
		kosten += kostenZondagFactor * Math.pow(dagenTussenZondagenInitieel + dagenTussenZondagen, kostenZondagMacht);
		kosten += kostenGebrokenDienstenFactor * Math.pow(dagenTussenGebrokenDienstenInitieel + dagenTussenGebrokenDiensten, kostenGebrokenDienstenMacht);
		kosten += kostenLangeDienstenFactor * Math.pow(dagenTussenLangeDienstenInitieel + dagenTussenLangeDiensten, kostenLangeDienstenMacht);
	}
	
	/**
	 * Berekent verandering in kosten als dienstA vervangen zou worden door dienstB
	 * @param dienstA
	 * @param regelA het regelnummer waarop dienstA zich bevindt
	 * @param dienstB
	 * @return de verandering van kosten
	 */
	private double verwisselKostenOverUren(Dienst dienstA, int regelA, Dienst dienstB)
	{
		int normaleWerkweekLengteInMinuten = (int)(60*rooster.getLengteWerkweek());
		double overMinutenA = Math.max(0, roosterInformatie.getWerkMinuten(regelA)-normaleWerkweekLengteInMinuten);
		double nieuweKosten = Math.max(0, overMinutenA-(dienstA.getDienstLengte().naarMinuten()
				-dienstB.getDienstLengte().naarMinuten())) * overUrenKosten;
		return nieuweKosten;
	}
	
	/**
	 * Berekent de verandering in kosten als dienstA wordt vervangen door dienstB.
	 * @param dienstA
	 * @param dienstB
	 * @param vorigeA de dienst die voor dienstA in de planning zit
	 * @param volgendeA de dienst die na dienstA in de planning zit
	 * @return de verandering van kosten
	 */
	private double verwisselKostenLaatVroeg(Dienst dienstA, Dienst dienstB, Dienst vorigeA, Dienst volgendeA)
	{
		double nieuweKosten = 0;
		if(dienstA.getDienstType() != dienstB.getDienstType())
		{
			if(dienstB.getDienstType()=='V')
			{
				if(vorigeA.getDienstType()=='V' && dienstA.getDienstType()=='L')
					nieuweKosten -= kostenLaatVroeg;
				else if(vorigeA.getDienstType()=='L' && dienstA.getDienstType()=='L')
					nieuweKosten += kostenLaatVroeg;
				if(volgendeA.getDienstType()=='V' && dienstA.getDienstType()=='L')
					nieuweKosten -= kostenLaatVroeg;
				else if(volgendeA.getDienstType()=='L' && dienstA.getDienstType()=='L')
					nieuweKosten += kostenLaatVroeg;
			}
			else if(dienstB.getDienstType()=='L')
			{
				if(vorigeA.getDienstType()=='L' && dienstA.getDienstType()=='V')
					nieuweKosten -= kostenLaatVroeg;
				else if(vorigeA.getDienstType()=='V' && dienstA.getDienstType()=='V')
					nieuweKosten += kostenLaatVroeg;
				if(volgendeA.getDienstType()=='L' && dienstA.getDienstType()=='V')
					nieuweKosten -= kostenLaatVroeg;
				else if(volgendeA.getDienstType()=='V' && dienstA.getDienstType()=='V')
					nieuweKosten += kostenLaatVroeg;
			}
		}
		return nieuweKosten;

	}
	
	/**
	 * Berekent de verandering in kosten voor de vroeg-vrij-vroeg en laat-vrij-laat situatie
	 * als dienstA wordt vervangen door dienstB.
	 * @param dienstA
	 * @param dienstB
	 * @param voorvorigeA de dienst gepland voor vorigeA
	 * @param vorigeA de dienst gepland voor dienstA
	 * @param volgendeA de dienst gepland na dienstA
	 * @param navolgendeA de dienst gepland na volgendeA
	 * @return
	 */
	private double verwisselKostenVroegVrijVroeg(Dienst dienstA, Dienst dienstB, 
			Dienst voorvorigeA, Dienst vorigeA, Dienst volgendeA, Dienst navolgendeA)
	{
		double nieuweKosten = 0;
		
		// nu kijken we gelijk naar vroeg-vrij-vroeg of laat-vrij-laat
		if(dienstB.isVrij())
		{
			if((vorigeA.getDienstType()=='V' && volgendeA.getDienstType()=='V')
				|| (vorigeA.getDienstType()=='L' && volgendeA.getDienstType()=='L'))
			{
				nieuweKosten += kostenVroegVrijVroeg;
			}
			if((voorvorigeA.getDienstType()=='V' && vorigeA.isVrij() && dienstA.getDienstType()=='V')
				|| (voorvorigeA.getDienstType()=='L' && vorigeA.isVrij() && dienstA.getDienstType()=='L'))
			{
				nieuweKosten -= kostenVroegVrijVroeg;
			}
			if((navolgendeA.getDienstType()=='V' && volgendeA.isVrij() && dienstA.getDienstType()=='V')
					|| (navolgendeA.getDienstType()=='L' && volgendeA.isVrij() && dienstA.getDienstType()=='L'))
			{
					nieuweKosten -= kostenVroegVrijVroeg;
			}
		}
		else if(dienstB.getDienstType()=='V')
		{
			if(dienstA.isVrij() && (vorigeA.getDienstType()=='V' || vorigeA.getDienstType()=='L')
					&& (volgendeA.getDienstType()=='V' || volgendeA.getDienstType()=='L'))
			{
				nieuweKosten -= kostenVroegVrijVroeg;
			}
			if(voorvorigeA.getDienstType()=='L' && vorigeA.isVrij())
				nieuweKosten += kostenVroegVrijVroeg;
			if(voorvorigeA.getDienstType()=='V' && vorigeA.isVrij() && dienstA.getDienstType()=='L')
				nieuweKosten -= kostenVroegVrijVroeg;
			if(navolgendeA.getDienstType()=='L' && volgendeA.isVrij())
				nieuweKosten += kostenVroegVrijVroeg;
			if(navolgendeA.getDienstType()=='V' && volgendeA.isVrij() && dienstA.getDienstType()=='L')
				nieuweKosten -= kostenVroegVrijVroeg;
		}
		else if(dienstB.getDienstType()=='L')
		{
			if(dienstA.isVrij() && (vorigeA.getDienstType()=='L' || vorigeA.getDienstType()=='V')
					&& (volgendeA.getDienstType()=='L' || volgendeA.getDienstType()=='V'))
			{
				nieuweKosten -= kostenVroegVrijVroeg;
			}
			if(voorvorigeA.getDienstType()=='V' && vorigeA.isVrij())
				nieuweKosten += kostenVroegVrijVroeg;
			if(voorvorigeA.getDienstType()=='L' && vorigeA.isVrij() && dienstA.getDienstType()=='V')
				nieuweKosten -= kostenVroegVrijVroeg;
			if(navolgendeA.getDienstType()=='V' && volgendeA.isVrij())
				nieuweKosten += kostenVroegVrijVroeg;
			if(navolgendeA.getDienstType()=='L' && volgendeA.isVrij() && dienstA.getDienstType()=='V')
				nieuweKosten -= kostenVroegVrijVroeg;
		}
		return nieuweKosten;
	}
	
	/**
	 * Berekent de verandering in kosten voor het werken op zaterdag/zondag als dienstA 
	 * wordt vervangen door dienstB.
	 * LET OP: er moet gelden dat dienstA en dienstB op zelfde dag zijn
	 * @param dienstA
	 * @param regelA regelnummer waarop dienstA zich bevindt
	 * @param dienstB
	 * @param kostenFactor factor in kostenfunctie
	 * @param kostenMacht macht in kostenfunctie
	 * @return de verandering in kosten
	 */
	private double verwisselKostenDag(Dienst dienstA, int regelA, Dienst dienstB, 
			double kostenFactor, int kostenMacht)
	{
		double nieuweKosten = 0;
		if(dienstA.getDienstType()!=dienstB.getDienstType())
		{
			int dag = dienstA.getDagNummer();
			boolean vorigeGevonden = false;
			boolean volgendeGevonden = false;
			int vorigeRegel = 0;
			int volgendeRegel = 0;
			int vorigeRegelTeller = regelA;
			int volgendeRegelTeller = regelA;
			
			while((!vorigeGevonden || !volgendeGevonden) && vorigeRegelTeller != regelA)
			{
				if(!dienstRooster[dag][vorigeRegelTeller].isVrij() && dienstRooster[dag][vorigeRegelTeller] != dienstA)
				{
					vorigeRegel = vorigeRegelTeller;
					vorigeGevonden = true;
				}
				if(!dienstRooster[dag][volgendeRegelTeller].isVrij() && dienstRooster[dag][volgendeRegelTeller] != dienstA)
				{
					volgendeRegel = volgendeRegelTeller;
					vorigeGevonden = true;
				}
				
				if(vorigeRegelTeller > 0)
					vorigeRegelTeller--;
				else
					vorigeRegelTeller = rooster.getRegels()-1;
				if(volgendeRegelTeller < rooster.getRegels()-1)
					volgendeRegelTeller++;
				else
					volgendeRegelTeller = 0;
			}
			// als we nu niet tevergeefs hebben gelopen, bewerk dan de oude kosten
			if(vorigeRegelTeller != regelA)
			{
				int dagenNaarVorigeDag;
				int dagenNaarVolgendeDag;
				if(vorigeRegel > regelA)
					dagenNaarVorigeDag = rooster.getRegels() - (vorigeRegel - regelA);
				else
					dagenNaarVorigeDag = regelA - vorigeRegel;
				if(volgendeRegel < regelA)
					dagenNaarVolgendeDag = rooster.getRegels() -(regelA - volgendeRegel);
				else
					dagenNaarVolgendeDag = volgendeRegel - regelA;
				//eerst het geval dat we nu vrij krijgen terwijl we eerst moesten werken
				if(dienstB.isVrij() && !dienstA.isVrij())
				{
					nieuweKosten -= kostenFactor * Math.pow(dagenNaarVorigeDag, kostenMacht);
					nieuweKosten -= kostenFactor * Math.pow(dagenNaarVolgendeDag, kostenMacht);
					nieuweKosten += kostenFactor * Math.pow(dagenNaarVorigeDag + dagenNaarVolgendeDag, kostenMacht);
				}
				// nu het geval dat we vrij hadden maar nu moeten werken
				else if (!dienstB.isVrij() && dienstA.isVrij())
				{
					nieuweKosten += kostenFactor * Math.pow(dagenNaarVorigeDag, kostenMacht);
					nieuweKosten += kostenFactor * Math.pow(dagenNaarVolgendeDag, kostenMacht);
					nieuweKosten -= kostenFactor * Math.pow(dagenNaarVorigeDag + dagenNaarVolgendeDag, kostenMacht);
				}
			}
		}
		return nieuweKosten;
	}
	
	/**
	 * Berekent de verandering in kosten als dienstA met dienstB vervangen zou worden voor 
	 * de component van de spreiding van gebroken diensten.
	 * @param dienstA
	 * @param regelA
	 * @param dienstB
	 * @return de verandering van kosten
	 */
	private double verwisselKostenGebrokenDiensten(Dienst dienstA, int regelA, Dienst dienstB)
	{
		double nieuweKosten = 0;
		if(dienstA.getDienstType()!=dienstB.getDienstType() && 
			(dienstA.getDienstType()=='G' || dienstB.getDienstType()=='G'))
		{
			boolean vorigeGevonden = false;
			boolean volgendeGevonden = false;
			int[] vor = vorigeDag(new int[] {dienstA.getDagNummer(), regelA});
			int[] vol = volgendeDag(new int[] {dienstA.getDagNummer(), regelA});
			Dienst vorige = dienstRooster[vor[0]][vor[1]];
			int vorigeDagenTeller = 0;
			Dienst volgende = dienstRooster[vol[0]][vol[1]];
			int volgendeDagenTeller = 0;
			
			while((!vorigeGevonden || !volgendeGevonden) && vorige != dienstA)
			{
				if(vorige.getDienstType()=='G')
					vorigeGevonden = true;
				else
					vorigeDagenTeller++;
				if(volgende.getDienstType()=='G')
					volgendeGevonden = true;
				else
					volgendeDagenTeller++;
				
				vor = vorigeDag(vor);
				vol = volgendeDag(vol);
				vorige = dienstRooster[vor[0]][vor[1]];
				volgende = dienstRooster[vol[0]][vol[1]];
			}
			// als we nu niet tevergeefs hebben gelopen, dan wis de oude kosten
			if(vorige != dienstA)
			{
				//eerst het geval dat we nu geen gebroken dienst krijgen terwijl we dat eerst wel hadden
				if(dienstA.getDienstType()=='G' /*dus dienstB is niet G*/ )
				{
					nieuweKosten -= kostenGebrokenDienstenFactor * Math.pow(vorigeDagenTeller, kostenGebrokenDienstenMacht);
					nieuweKosten -= kostenGebrokenDienstenFactor * Math.pow(volgendeDagenTeller, kostenGebrokenDienstenMacht);
					nieuweKosten += kostenGebrokenDienstenFactor * Math.pow(vorigeDagenTeller + volgendeDagenTeller, kostenGebrokenDienstenMacht);
				}
				// nu het geval dat we een gebroken dienst krijgen hadden maar dat eerst niet hadden
				else if (dienstB.getDienstType()=='G')
				{
					nieuweKosten += kostenGebrokenDienstenFactor * Math.pow(vorigeDagenTeller, kostenGebrokenDienstenMacht);
					nieuweKosten += kostenGebrokenDienstenFactor * Math.pow(volgendeDagenTeller, kostenGebrokenDienstenMacht);
					nieuweKosten -= kostenGebrokenDienstenFactor * Math.pow(vorigeDagenTeller + volgendeDagenTeller, kostenGebrokenDienstenMacht);
				}
			}
		}
		return nieuweKosten;
	}
	
	/**
	 * Berekent de verandering in kosten als dienstA met dienstB vervangen zou worden.
	 * @param dienstA
	 * @param regelA
	 * @param dienstB
	 * @return de verandering van kosten
	 */
	private double verwisselKostenLangeDiensten(Dienst dienstA, int regelA, Dienst dienstB)
	{
		double nieuweKosten = 0;
		boolean aLang = dienstA.getDienstLengte().naarMinuten() > 510;
		boolean bLang = dienstB.getDienstLengte().naarMinuten() > 510;
		if((aLang || bLang) && (!aLang && bLang))
		{
			boolean vorigeGevonden = false;
			boolean volgendeGevonden = false;
			int[] vor = vorigeDag(new int[] {dienstA.getDagNummer(), regelA});
			int[] vol = volgendeDag(new int[] {dienstA.getDagNummer(), regelA});
			Dienst vorige = dienstRooster[vor[0]][vor[1]];
			int vorigeDagenTeller = 0;
			Dienst volgende = dienstRooster[vol[0]][vol[1]];
			int volgendeDagenTeller = 0;
			
			while((!vorigeGevonden || !volgendeGevonden) && vorige != dienstA)
			{
				if(vorige.getDienstLengte().naarMinuten() > 510)
					vorigeGevonden = true;
				else
					vorigeDagenTeller++;
				if(volgende.getDienstLengte().naarMinuten() > 510)
					volgendeGevonden = true;
				else
					volgendeDagenTeller++;
				
				vor = vorigeDag(vor);
				vol = volgendeDag(vol);
				vorige = dienstRooster[vor[0]][vor[1]];
				volgende = dienstRooster[vol[0]][vol[1]];
			}
			// als we nu niet tevergeefs hebben gelopen, dan wis de oude kosten
			if(vorige != dienstA)
			{
				//eerst het geval dat we nu geen lange dienst krijgen terwijl we dat eerst wel hadden
				if(bLang)
				{
					nieuweKosten -= kostenLangeDienstenFactor * Math.pow(vorigeDagenTeller, kostenLangeDienstenMacht);
					nieuweKosten -= kostenLangeDienstenFactor * Math.pow(volgendeDagenTeller, kostenLangeDienstenMacht);
					nieuweKosten += kostenLangeDienstenFactor * Math.pow(vorigeDagenTeller + volgendeDagenTeller, kostenLangeDienstenMacht);
				}
				// nu het geval dat we een gebroken dienst krijgen hadden maar dat eerst niet hadden
				else if (aLang)
				{
					nieuweKosten += kostenLangeDienstenFactor * Math.pow(vorigeDagenTeller, kostenLangeDienstenMacht);
					nieuweKosten += kostenLangeDienstenFactor * Math.pow(volgendeDagenTeller, kostenLangeDienstenMacht);
					nieuweKosten -= kostenLangeDienstenFactor * Math.pow(vorigeDagenTeller + volgendeDagenTeller, kostenLangeDienstenMacht);
				}
			}
		}
		return nieuweKosten;
	}
	
	/**
	 * Berekent de verandering in kosten als dienstA wordt vervangen door dienstB.
	 * LET OP: dienstA en dienstB moeten zich op dezelfde dag bevinden
	 * @param dienstA
	 * @param regelA
	 * @param dienstB
	 * @param vorigeA
	 * @param volgendeA
	 * @return verandering van de kosten
	 */
	private double verwisselKostenZaterdagVrijZondagWerken(Dienst dienstA, int regelA, 
			Dienst dienstB, Dienst vorigeA, Dienst volgendeA)
	{
		double nieuweKosten = 0;
		if(dienstA.isVrij() != dienstB.isVrij())
		{			
			if(dienstA.getDagNummer()==6)
			{
				if(dienstA.isVrij() && !volgendeA.isVrij())
					nieuweKosten -= kostenZaterdagVrijZondagWerken;
				else if(dienstB.isVrij() && !volgendeA.isVrij())
					nieuweKosten += kostenZaterdagVrijZondagWerken;
			}
			else if(dienstA.getDagNummer()==0)
			{
				if(vorigeA.isVrij() && !dienstA.isVrij())
					nieuweKosten -= kostenZaterdagVrijZondagWerken;
				else if(vorigeA.isVrij() && !dienstB.isVrij())
					nieuweKosten += kostenZaterdagVrijZondagWerken;
			}
		}
		return nieuweKosten;
	}
	
	/**
	 * Berekent verandering van kosten bij vervanging van dienstA door dienstB en ook de 
	 * nieuwe wekelijkse rust.
	 * @param dienstA
	 * @param regelA
	 * @param dienstB
	 * @return [verandering van kosten, nieuwe wekelijkse rust]
	 */
	private double[] verwisselKostenAaneengeslotenRust(Dienst dienstA, int regelA, Dienst dienstB)
	{
		double nieuweKosten = 0;
		double wekelijkseRust = roosterInformatie.getWekelijkseRust(regelA);
		Tijdstip middernacht = new Tijdstip(24, 00);
		if(dienstB.isVrij())
		{
			boolean vorigeGevonden = false;
			boolean volgendeGevonden = false;
			int minutenRust = 60;
			int[] vor = vorigeDag(new int[] {dienstA.getDagNummer(), regelA});
			int[] vol = volgendeDag(new int[] {dienstA.getDagNummer(), regelA});
			while((!vorigeGevonden || !volgendeGevonden) && dienstRooster[vor[0]][vor[1]]!=dienstA)
			{
				if(dienstRooster[vor[0]][vor[1]].isVrij())
				{
					minutenRust += 60;
					vor = vorigeDag(vor);
				}
				else
				{
					minutenRust += (middernacht.minus(dienstRooster[vor[0]][vor[1]].getEindTijd())).naarMinuten();
					vorigeGevonden = true;
				}
				if(dienstRooster[vol[0]][vol[1]].isVrij())
				{
					minutenRust += 60;
					vol = volgendeDag(vol);
				}
				else
				{
					minutenRust += (dienstRooster[vol[0]][vol[1]].getBeginTijd().minus(middernacht)).naarMinuten();
					volgendeGevonden = true;
				}
			}
			if(minutenRust > roosterInformatie.getWekelijkseRust(regelA))
			{
				nieuweKosten += kostenAaneengeslotenRustFactor * Math.pow(((float)minutenRust)/60.0, 
						kostenAaneengeslotenRustMacht);
				nieuweKosten -= kostenAaneengeslotenRustFactor * Math.pow(((float)roosterInformatie.getWekelijkseRust(regelA))/60.0, 
						kostenAaneengeslotenRustMacht);
				wekelijkseRust = ((float)minutenRust)/60.0;
			}
		}
		else 
		// we stoppen nu wellicht een werkdag middenin wat voorheen de aaneengesloten rust was, dus we moeten de aaneengesloten rust opnieuw bepalen
		{
			int minutenRust = 0;
			int maxMinutenRust = 0;
			boolean klaar = false;
			int volgendeWeken = 0; // hoeveel weken er verder in het rooster wordt gekeken
			int[] vol = new int[] {0, regelA};
			while(!klaar && volgendeWeken < 2)
			{
				int dag = vol[0];
				Dienst volgendeDienst = dienstRooster[dag][vol[1]];
				if(!volgendeDienst.isVrij())
				{
					minutenRust += (volgendeDienst.getBeginTijd().minus(middernacht)).naarMinuten();
					maxMinutenRust = Math.max(minutenRust, maxMinutenRust);
					minutenRust = (middernacht.minus(volgendeDienst.getEindTijd())).naarMinuten();
				}
				else
					minutenRust += 1440; //60*24
				if(dag >= 6)
				{
					if(volgendeDienst.isVrij())
					{
						volgendeWeken++;
					}
					else
						klaar = true;
				}
				vol = volgendeDag(vol);
			}
			
			nieuweKosten -= kostenAaneengeslotenRustFactor * Math.pow(((float)roosterInformatie.getWekelijkseRust(regelA))/60.0, 
					kostenAaneengeslotenRustMacht);
			nieuweKosten += kostenAaneengeslotenRustFactor * Math.pow(((float)maxMinutenRust)/60.0, 
					kostenAaneengeslotenRustMacht);
			wekelijkseRust = ((float)maxMinutenRust)/60.0;
		}
		return new double[] {nieuweKosten, wekelijkseRust};
	}
	
	private double verwisselKostenTeWeinigTijdTussenDiensten(Dienst dienstA, Dienst dienstB, 
			Dienst vorigeA, Dienst volgendeA)
	{
		double nieuweKosten = 0;
		// eerst eventuele kosten wegwerken
		if(!dienstA.isVrij())
		{
			if(!vorigeA.isVrij())
			{
				if((dienstA.getEindTijd().minus(vorigeA.getEindTijd())).naarMinuten()<660) //11*60
					nieuweKosten -= kostenTeWeinigRustTussenDiensten;
			}
			if(!volgendeA.isVrij())
			{
				if((volgendeA.getEindTijd().minus(dienstA.getEindTijd())).naarMinuten()<660) //11*60
					nieuweKosten -= kostenTeWeinigRustTussenDiensten;
			}
		}
		if(!dienstB.isVrij())
		{
			if(!vorigeA.isVrij())
			{
				if((dienstB.getEindTijd().minus(vorigeA.getEindTijd())).naarMinuten()<660) //11*60
					nieuweKosten += kostenTeWeinigRustTussenDiensten;
			}
			if(!volgendeA.isVrij())
			{
				if((volgendeA.getEindTijd().minus(dienstB.getEindTijd())).naarMinuten()<660) //11*60
					nieuweKosten += kostenTeWeinigRustTussenDiensten;
			}
		}
		return nieuweKosten;
	}
	
	private double verwisselKostenTeWeinigWekelijkseRust(Dienst dienstA, int regelA, Dienst dienstB, double rustUren)
	{
		double nieuweKosten = 0;
		// ten opzichte van vorige week
		int vorigeRegel;
		if(regelA>0)
			vorigeRegel = regelA;
		else
			vorigeRegel = 0;
	
		if(rustUren < 36)
			nieuweKosten += kostenTeWeinigWekelijkseRust;
		double vorigeWeekTekort = Math.max(0, 45-roosterInformatie.getWekelijkseRust(vorigeRegel));
		if(vorigeWeekTekort>0 && rustUren-vorigeWeekTekort<45)
			nieuweKosten += kostenTeWeinigWekelijkseRust;
		
		// ten opzichte van volgende week
		int volgendeRegel;
		if(regelA>0)
			volgendeRegel = regelA;
		else
			volgendeRegel = 0;
		double volgendeWeekTekort = Math.max(0, 45-rustUren);
		if(volgendeWeekTekort>0 && roosterInformatie.getWekelijkseRust(volgendeRegel)-volgendeWeekTekort<45)
			nieuweKosten += kostenTeWeinigWekelijkseRust;
		return nieuweKosten;
	}
	
	/**
	 * Berekent de nieuwe kosten van dit dienstrooster ALS dienstA en dienstB verwisseld
	 * zouden worden. De verwisseling wordt dus nog niet uitgevoerd.
	 * @param dagA dagcomponent waarmee dienstA in het dienstRooster kan worden opgezocht
	 * @param regelA regelcomponent waarmee dienstA in het dienstRooster kan worden opgezocht
	 * @param dagB dagcomponent waarmee dienstB in het dienstRooster kan worden opgezocht
	 * @param regelB regelcomponent waarmee dienstB in het dienstRooster kan worden opgezocht
	 * @return
	 */
	public double berekenKostenOPTintern(int dagA, int regelA, int dagB, int regelB)
	{
		double nieuweKosten = getKosten();
		Dienst dienstA = dienstRooster[dagA][regelA];
		Dienst dienstB = dienstRooster[dagB][regelB];
		
		// bereken de kosten voor overuren
		nieuweKosten += verwisselKostenOverUren(dienstA, regelA, dienstB);
		nieuweKosten += verwisselKostenOverUren(dienstB, regelB, dienstA);
		/* we hoeven de kosten voor teveel werkdagen niet opnieuw te bereken, omdat we binnen
		 * hetzelfde rooster veranderen
		 */
		
		int[] vorA = vorigeDag(new int[] {dagA, regelA});
		int[] volA = volgendeDag(new int[] {dagA, regelA});
		int[] vorB = vorigeDag(new int[] {dagB, regelB});
		int[] volB = volgendeDag(new int[] {dagB, regelB});
		Dienst vorigeA = dienstRooster[vorA[0]][vorA[1]];
		Dienst volgendeA = dienstRooster[volA[0]][volA[1]];
		Dienst vorigeB = dienstRooster[vorB[0]][vorB[1]];
		Dienst volgendeB = dienstRooster[volB[0]][volB[1]];
		int[] vorvorA = vorigeDag(vorA);
		int[] volvolA = volgendeDag(volA);
		int[] vorvorB = vorigeDag(vorB);
		int[] volvolB = volgendeDag(volB);
		Dienst voorvorigeA = dienstRooster[vorvorA[0]][vorvorA[1]];
		Dienst navolgendeA = dienstRooster[volvolA[0]][volvolA[1]];
		Dienst voorvorigeB = dienstRooster[vorvorB[0]][vorvorB[1]];
		Dienst navolgendeB = dienstRooster[volvolB[0]][volvolB[1]];
		
		// nu kijken of er sprake was van een vroeg-laat of laat-vroeg situatie
		nieuweKosten += verwisselKostenLaatVroeg(dienstA, dienstB, vorigeA, volgendeA);
		nieuweKosten += verwisselKostenLaatVroeg(dienstB, dienstA, vorigeB, volgendeB);
		
		// en of er sprake is van vroeg-vrij-vroeg of laat-vrij-laat
		nieuweKosten += verwisselKostenVroegVrijVroeg(dienstA, dienstB, voorvorigeA, vorigeA, volgendeA, navolgendeA);
		nieuweKosten += verwisselKostenVroegVrijVroeg(dienstB, dienstA, voorvorigeB, vorigeB, volgendeB, navolgendeB);
		
		// nu de boete voor de spreiding van gewerkte zaterdagen
		if(dienstA.getDagNummer()==6)
		{
			nieuweKosten += verwisselKostenDag(dienstA, regelA, dienstB, kostenZaterdagFactor, kostenZaterdagMacht);
			nieuweKosten += verwisselKostenDag(dienstB, regelB, dienstA, kostenZaterdagFactor, kostenZaterdagMacht);
		}
		// en de boete voor de spreiding van gewerkte zondagen
		else if(dienstA.getDagNummer()==0)
		{
			nieuweKosten += verwisselKostenDag(dienstA, regelA, dienstB, kostenZondagFactor, kostenZondagMacht);
			nieuweKosten += verwisselKostenDag(dienstB, regelB, dienstA, kostenZondagFactor, kostenZondagMacht);
		}
		// nu de boete voor de spreiding van gebroken diensten
		nieuweKosten += verwisselKostenGebrokenDiensten(dienstA, regelA, dienstB);
		nieuweKosten += verwisselKostenGebrokenDiensten(dienstB, regelB, dienstA);
		
		// en de boete voor de spreiding van lange diensten
		nieuweKosten += verwisselKostenLangeDiensten(dienstA, regelA, dienstB);
		nieuweKosten += verwisselKostenLangeDiensten(dienstB, regelB, dienstA);
		
		// de boete voor de weekendrust
		double[] rustA = verwisselKostenAaneengeslotenRust(dienstA, regelA, dienstB);
		double[] rustB = verwisselKostenAaneengeslotenRust(dienstB, regelB, dienstA);
		nieuweKosten += rustA[0] + rustB[0];
		
		// de boete voor zaterdag vrij en zondag werken
		nieuweKosten += verwisselKostenZaterdagVrijZondagWerken(dienstA, regelA, dienstB, vorigeA, volgendeA);
		nieuweKosten += verwisselKostenZaterdagVrijZondagWerken(dienstB, regelB, dienstA, vorigeA, volgendeA);
		
		// boete voor te weinig rust tussen twee opvolgende diensten
		nieuweKosten += verwisselKostenTeWeinigTijdTussenDiensten(dienstA, dienstB, vorigeA, volgendeA);
		nieuweKosten += verwisselKostenTeWeinigTijdTussenDiensten(dienstB, dienstA, vorigeB, volgendeB);
		
		// boete voor te weinig wekelijkse rust
		nieuweKosten += verwisselKostenTeWeinigWekelijkseRust(dienstA, regelA, dienstB, rustA[1]);
		nieuweKosten += verwisselKostenTeWeinigWekelijkseRust(dienstB, regelB, dienstA, rustB[1]);
		
		return nieuweKosten;
	}
	
	public double berekenKostenOPTextern(int dagA, int regelA, Dienst dienstB)
	{
		double nieuweKosten = getKosten();
		Dienst dienstA = dienstRooster[dagA][regelA];
		
		// bereken de kosten voor overuren
		nieuweKosten += verwisselKostenOverUren(dienstA, regelA, dienstB);
		
		// kosten voor teveel werkdagen
		if(dienstA.isVrij() && !dienstB.isVrij())
		{
			nieuweKosten -= Math.max(0, roosterInformatie.getWerkdagen()-rooster.getWerkdagen()) * extraWerkDagenKosten;
			nieuweKosten += Math.max(0, roosterInformatie.getWerkdagen()+1-rooster.getWerkdagen()) * extraWerkDagenKosten;
		}
		else if(!dienstA.isVrij() && dienstB.isVrij())
		{
			nieuweKosten -= Math.max(0, roosterInformatie.getWerkdagen()-rooster.getWerkdagen()) * extraWerkDagenKosten;
			nieuweKosten += Math.max(0, roosterInformatie.getWerkdagen()-1-rooster.getWerkdagen()) * extraWerkDagenKosten;
		}
		
		int[] vorA = vorigeDag(new int[] {dagA, regelA});
		int[] volA = volgendeDag(new int[] {dagA, regelA});
		Dienst vorigeA = dienstRooster[vorA[0]][vorA[1]];
		Dienst volgendeA = dienstRooster[volA[0]][volA[1]];
		int[] vorvorA = vorigeDag(vorA);
		int[] volvolA = volgendeDag(volA);
		Dienst voorvorigeA = dienstRooster[vorvorA[0]][vorvorA[1]];
		Dienst navolgendeA = dienstRooster[volvolA[0]][volvolA[1]];
		
		// nu kijken of er sprake was van een vroeg-laat of laat-vroeg situatie
		nieuweKosten += verwisselKostenLaatVroeg(dienstA, dienstB, vorigeA, volgendeA);
		
		// en of er sprake is van vroeg-vrij-vroeg of laat-vrij-laat
		nieuweKosten += verwisselKostenVroegVrijVroeg(dienstA, dienstB, voorvorigeA, vorigeA, volgendeA, navolgendeA);
		
		// nu de boete voor de spreiding van gewerkte zaterdagen
		if(dienstA.getDagNummer()==6)
		{
			nieuweKosten += verwisselKostenDag(dienstA, regelA, dienstB, kostenZaterdagFactor, kostenZaterdagMacht);
		}
		// en de boete voor de spreiding van gewerkte zondagen
		else if(dienstA.getDagNummer()==0)
		{
			nieuweKosten += verwisselKostenDag(dienstA, regelA, dienstB, kostenZondagFactor, kostenZondagMacht);
		}
		// nu de boete voor de spreiding van gebroken diensten
		nieuweKosten += verwisselKostenGebrokenDiensten(dienstA, regelA, dienstB);
		
		// en de boete voor de spreiding van lange diensten
		nieuweKosten += verwisselKostenLangeDiensten(dienstA, regelA, dienstB);
		
		// de boete voor de weekendrust
		double[] rustA = verwisselKostenAaneengeslotenRust(dienstA, regelA, dienstB);
		nieuweKosten += rustA[0];
		
		// de boete voor zaterdag vrij en zondag werken
		nieuweKosten += verwisselKostenZaterdagVrijZondagWerken(dienstA, regelA, dienstB, vorigeA, volgendeA);
		
		// kosten voor te weinig rusttijd tussen opvolgende diensten
		nieuweKosten += verwisselKostenTeWeinigTijdTussenDiensten(dienstA, dienstB, vorigeA, volgendeA);
		
		// boete voor te weinig wekelijkse rust
		nieuweKosten += verwisselKostenTeWeinigWekelijkseRust(dienstA, regelA, dienstB, rustA[1]);
		
		return nieuweKosten;
	}
	
	/**
	 * Hulpmethode die in het gegeven dienstRooster de volgende dag oplevert.
	 * @param huidigeDag, gegeven als [dag, regelnummer]
	 * @return de volgende dag, ook gegeven als [dag, regelnummer]
	 */
	private int[] volgendeDag(int[] huidigeDag)
	{
		int dag = huidigeDag[0];
		int regel = huidigeDag[1];
		
		if(dag < 6)
			return new int[] {dag+1, regel};
		else
		{
			if(regel < rooster.getRegels()-1)
				return new int[] {0, regel+1};
			else
				return new int[] {0, 0};
		}
	}
	
	/**
	 * Hulpmethode die in het gegeven dienstRooster de vorige dag oplevert.
	 * @param huidigeDag, gegeven als [dag, regelnummer]
	 * @return de vorige dag, ook gegeven als [dag, regelnummer]
	 */
	private int[] vorigeDag(int[] huidigeDag)
	{
		int dag = huidigeDag[0];
		int regel = huidigeDag[1];
		
		if(dag > 0)
			return new int[] {dag-1, regel};
		else
		{
			if(regel > 0)
				return new int[] {6, regel-1};
			else
				return new int[] {6, rooster.getRegels()-1};
		}
	}
	
	public void addDienst(int regelNummer, int dagNummer, Dienst dienst)
	{
		dienstRooster[dagNummer][regelNummer] = dienst;
	}

	public Dienst getDienst(int regelNummer, int dagNummer)
	{
		return dienstRooster[dagNummer][regelNummer];
	}

	public int getAantalRegels() 
	{
		return dienstRooster[1].length;
	}

	public double getKosten() 
	{
		return kosten;
	}

	public int getRoosterNr() 
	{
		return roosterNr;
	}

	public void setRoosterNr(int roosterNr) 
	{
		this.roosterNr = roosterNr;
	}

	public Dienst[][] getDienstRooster() {
		return dienstRooster;
	}
}
